"""
See https://en.wikipedia.org/wiki/Futoshiki

Example of Execution:
  python3 Futoshiki.py -data=Futoshiki_futo3_0.json
"""

from pycsp3 import *

n, nbHints, opHints = data  # n is the order of the grid

# x[i][j] is the number put at row i and column j
x = VarArray(size=[n, n], dom=range(1, n + 1))

satisfy(
    # different values on each row and each column
    AllDifferent(x, matrix=True),

    # respecting number hints
    [x[i][j] == k for (i, j, k) in nbHints],

    # respecting operator hints
    [y < z if lt else y > z for (y, z, lt) in [(x[i][j], x[i][j + 1] if hr else x[i + 1][j], lt) for (i, j, lt, hr) in opHints]]
)



"""
See Problem in Minizinc (Challenge 2012)

Examples of Execution:
  python3 Fastfood.py -data=Fastfood_ff01.json
  python3 Fastfood.py -data=Fastfood_ff01.json -variant=table
  python3 Fastfood.py -data=Fastfood_example.dzn -dataparser=Fastfood_ParserZ.py
"""

from pycsp3 import *

nDepots, restaurants = data
nRestaurants = len(restaurants)
positions = [restaurant.position for restaurant in restaurants]
# NOTE: below, cp_array is necessary for being able to use the constraint Element in the main variant
distances = cp_array([abs(positions[j] - positions[i]) for j in range(nRestaurants)] for i in range(nRestaurants))

# d[i][j] is the distance between the ith restaurant and the jth depot
d = VarArray(size=[nRestaurants, nDepots], dom=lambda i, j: distances[i])

if not variant():
    # x[j] is the index of the restaurant used as the jth depot
    x = VarArray(size=nDepots, dom=range(nRestaurants))

    satisfy(
        # linking positions of depots with their distances to the restaurants
        distances[i][x[j]] == d[i][j] for i in range(nRestaurants) for j in range(nDepots)
    )

elif variant("table"):
    # x[j] is the position of the jth depot
    x = VarArray(size=nDepots, dom=positions)

    satisfy(
        # linking positions of depots with their distances to the restaurants
        (x[j], d[i][j]) in {(p, abs(p - positions[i])) for p in positions} for i in range(nRestaurants) for j in range(nDepots)
    )

satisfy(
    # tag(symmetry-breaking)
    Increasing(x, strict=True)
)

minimize(
    Sum(Minimum(d[i]) for i in range(nRestaurants))
)



"""
The Zebra puzzle (sometimes referred to as Einstein's puzzle) is defined as follows.
There are five houses in a row, numbered from left to right.
Each of the five houses is painted a different color, and has one inhabitant.
The inhabitants are all of different nationalities, own different pets, drink different beverages and have different jobs.
We know that:
 - colors are yellow, green, red, white, and blue
 - nations of inhabitants are italy, spain, japan, england, and norway
 - pets are cat, zebra, bear, snails, and horse
 - drinks are milk, water, tea, coffee, and juice
 - jobs are painter, sculptor, diplomat, pianist, and doctor
 - the painter owns the horse
 - the diplomat drinks coffee
 - the one who drinks milk lives in the white house
 - the Spaniard is a painter
 - the Englishman lives in the red house
 - the snails are owned by the sculptor
 - the green house is on the left of the red one
 - the Norwegian lives on the right of the blue house
 - the doctor drinks milk
 - the diplomat is Japanese
 - the Norwegian owns the zebra
 - the green house is next to the white one
 - the horse is owned by the neighbor of the diplomat
 - the Italian either lives in the red, white or green house

Execution:
  python3 Zebra.py
"""

from pycsp3 import *

houses = range(5)  # each house has a number from 0 (left) to 4 (right)

# colors[i] is the house of the ith color
yellow, green, red, white, blue = colors = VarArray(size=5, dom=houses)

# nations[i] is the house of the inhabitant with the ith nationality
italy, spain, japan, england, norway = nations = VarArray(size=5, dom=houses)

# jobs[i] is the house of the inhabitant with the ith job
painter, sculptor, diplomat, pianist, doctor = jobs = VarArray(size=5, dom=houses)

# pets[i] is the house of the inhabitant with the ith pet
cat, zebra, bear, snails, horse = pets = VarArray(size=5, dom=houses)

# drinks[i] is the house of the inhabitant with the ith preferred drink
milk, water, tea, coffee, juice = drinks = VarArray(size=5, dom=houses)

satisfy(
    AllDifferent(colors),
    AllDifferent(nations),
    AllDifferent(jobs),
    AllDifferent(pets),
    AllDifferent(drinks),

    painter == horse,
    diplomat == coffee,
    white == milk,
    spain == painter,
    england == red,
    snails == sculptor,
    green + 1 == red,
    blue + 1 == norway,
    doctor == milk,
    japan == diplomat,
    norway == zebra,
    abs(green - white) == 1,
    horse in {diplomat - 1, diplomat + 1},
    italy in {red, white, green}
)



"""
See http://mathworld.wolfram.com/deBruijnSequence.html and http://www.hakank.org/comb/debruijn.cgi#info

Examples of Execution:
  python3 DeBruijnSequence.py
  python3 DeBruijnSequence.py -data=[2,5]
"""

from pycsp3 import *

b, n = data or (3, 4)
m = b ** n
powers = [b ** i for i in range(n - 1, -1, -1)]

# x[i] is the ith number (in base 10) of the sequence
x = VarArray(size=m, dom=range(m))

# d[i][j] is the jth digit in the base for the ith number of the sequence; d[i] is then the representation in the base of the ith number
d = VarArray(size=[m, n], dom=range(b))

# g[i] is the number of occurrences of digit i in the first column of array d
g = VarArray(size=b, dom=range(m + 1))

satisfy(
    AllDifferent(x),

    # linking x and d: d[i] is the representation of x[i] in base 2
    [d[i] * powers == x[i] for i in range(m)],

    # de Bruijn condition
    [d[i - 1][j] == d[i % m][j - 1] for i in range(1, m + 1) for j in range(1, n)],

    Cardinality(d[:, 0], occurrences={i: g[i] for i in range(b)}),

    # tag(symmetry-breaking)
    Minimum(x) == x[0]
)



from pycsp3 import *

n, d = data  # number of dominoes -- number of values

# x[i] is the value of the ith domino
x = VarArray(size=n, dom=range(d))

if not variant():
    satisfy(
        AllEqual(x),

        (x[0] + 1 == x[- 1]) | ((x[0] == x[- 1]) & (x[0] == d - 1))
    )

elif variant("table"):
    satisfy(
        [(x[i], x[i + 1]) in {(v, v) for v in range(d)} for i in range(n - 1)],

        (x[0], x[- 1]) in {(v + 1, v) for v in range(d - 1)} | {(d - 1, d - 1)}
    )


# TODO: would it be possible to extend the code so as be able to to write : x[0] == x[- 1] == v - 1? this is not obvious


import random

from pycsp3.compiler import Compilation
from pycsp3.problems.data.parsing import *

nCardsPerSuit = ask_number("Number of cards per suit (e.g., 13)")
nCardsPerPile = ask_number("Number of cards per pile (e.g., 3)")
seed = ask_number("Seed")

nCards = 4 * nCardsPerSuit
nPiles = (nCards - 1) // nCardsPerPile
assert (nCards - 1) % nCardsPerPile == 0

random.seed(seed)
cards = list(range(nCards))
piles = []
for i in range(nPiles):
    pile = []
    for j in range(nCardsPerPile):
        v = random.randrange(len(cards))
        pile.append(cards[v])
        del cards[v]
    piles.append(pile)

data["nCardsPerSuit"] = nCardsPerSuit
data["piles"] = piles

Compilation.string_data = "-" + "-".join(str(v) for v in (nCardsPerSuit, seed))

# TODO not a good way of building instances because they are most of the time  trivially unsatisfiable



"""
See Problem in MiniZinc -- https://github.com/MiniZinc/minizinc-benchmarks/tree/master/league

Make leagues for some games.
 - Ranking should be close in each league.
 - In a league, variety of country (where player comes from) is needed.

Example of Execution:
  python3 League.py -data=League_010-03-04.json
"""

from pycsp3 import *

leagueSize, players = data  # rankings and countries of players
rankings, countries = zip(*players)
nPlayers = len(players)
nLeagues = nPlayers // leagueSize + (1 if nPlayers % leagueSize != 0 else 0)
nFullLeagues = nLeagues if nPlayers % leagueSize == 0 else nLeagues - (leagueSize - nPlayers % leagueSize)
sizes = [leagueSize + (0 if i < nFullLeagues else -1) for i in range(nLeagues)]

# p[i][j] is the jth player of the ith league
p = VarArray(size=[nLeagues, leagueSize], dom=lambda i, j: range(nPlayers) if j < sizes[i] else None)

# r[i][j] is the ranking of the jth player of the ith league
r = VarArray(size=[nLeagues, leagueSize], dom=lambda i, j: rankings if j < sizes[i] else None)

# c[i][j] is the country of the jth player of the ith league
c = VarArray(size=[nLeagues, leagueSize], dom=lambda i, j: countries if j < sizes[i] else None)

table = {(i, rankings[i], countries[i]) for i in range(nPlayers)}

satisfy(
    # each player belongs to only one league
    AllDifferent(p),

    # linking players with their rankings and countries
    [(p[i][j], r[i][j], c[i][j]) in table for i in range(nLeagues) for j in range(sizes[i])]
)

minimize(
    # minimizing overall differences between highest and lowest rankings of players in leagues while paying attention to numbers of countries
    Sum(Maximum(r[i]) - Minimum(r[i]) for i in range(nLeagues)) * 100 - Sum(NValues(c[i]) for i in range(nLeagues))
)



from pycsp3 import *
from pycsp3.tools.utilities import value_in_base

nDisks = data
nTowers = 3
nStates, nSteps = nTowers ** nDisks, 2 ** nDisks - 1

x = VarArray(size=nSteps - 1, dom=range(nStates))


def table():
    def are_compatible(state1, state2):
        t1, t2 = value_in_base(state1, nDisks, nTowers), value_in_base(state2, nDisks, nTowers)
        frozen_towers = [False] * nTowers
        for i in range(len(t1) - 1, -1, -1):
            if t1[i] != t2[i]:
                break
            frozen_towers[t1[i]] = True
        return i >= 0 and not frozen_towers[t1[i]] and not frozen_towers[t2[i]] and all(t1[j] == t2[j] for j in range(i))

    return {(i, j) for i in range(nStates) for j in range(nStates) if are_compatible(i, j)}


table = table()

satisfy(
    x[0] in {1, 2},

    [(x[i], x[i + 1]) in table for i in range(nSteps - 2)]
)



"""
See https://en.wikipedia.org/wiki/Verbal_arithmetic

A model for a general form of this problem is in CryptoPuzzle.py

Execution:
  python3 SendMore.py
"""

from pycsp3 import *

# letters[i] is the digit of the ith letter involved in the equation
s, e, n, d, m, o, r, y = letters = VarArray(size=8, dom=range(10))

satisfy(
    # letters are given different values
    AllDifferent(letters),

    # words cannot start with 0
    [s > 0, m > 0],

    # respecting the mathematical equation
    [s, e, n, d] * [1000, 100, 10, 1] + [m, o, r, e] * [1000, 100, 10, 1] == [m, o, n, e, y] * [10000, 1000, 100, 10, 1]
)


# some other possible ways of posting the sum:
# [m, o - s - m, n - e - o, e - n - r, y - d - e] * [10000, 1000, 100, 10, 1] == 0
# [s + m, e + o, n + r, d + e] * [1000, 100, 10, 1] == [m, o, n, e, y] * [10000, 1000, 100, 10, 1]



import re

from pycsp3.classes.auxiliary.ptypes import TypeVar
from pycsp3.classes.main.domains import Domain


class NotVariable:
    def __init__(self, variable):
        self.variable = variable


class NegVariable:
    def __init__(self, variable):
        self.variable = variable


class Variable:
    name2obj = dict()  # Dictionary (keys: names of variables - values: variable objects)

    @staticmethod
    def build_names_array(name, sizes, mins, indexes=[]):
        if sizes:
            t = []
            for i in range(sizes[0]):
                indexes.append(i + mins[len(indexes)])
                t.append(Variable.build_names_array(name, sizes[1:], mins, indexes))
                indexes.pop()
            return t
        return name + "[" + "][".join(str(i) for i in indexes) + "]"

    @staticmethod
    def build_domain(name, domain, indexes):
        if domain is None:
            return None
        if isinstance(domain, Domain):
            return domain
        if isinstance(domain, type(lambda: 0)):
            domain = domain(*indexes)
            if domain is None:
                return None
        if isinstance(domain, (tuple, list)):
            if all(isinstance(v, int) for v in domain) or all(isinstance(v, str) for v in domain):  # possible, even if using a set is recommended
                return Domain(set(domain))
            # at this point, it means that a specific domain for each variable is given in a list
            for i in indexes:
                assert i < len(domain), "The number of domains is less than the specified index " + name + " - " + str(domain)
                domain = domain[i]
        if isinstance(domain, Domain):
            return domain
        if isinstance(domain, list) and (all(isinstance(v, int) for v in domain) or all(isinstance(v, str) for v in domain)):
            return Domain(set(domain))
        assert isinstance(domain, (range, set)), type(domain)
        return Domain(domain)

    @staticmethod
    def build_variable(name, domain, indexes):
        dom = Variable.build_domain(name, domain, indexes)
        if dom is None:
            return None
        var = VariableInteger(name, dom) if dom.get_type() == TypeVar.INTEGER else VariableSymbolic(name, dom)
        Variable.name2obj[name] = var
        return var

    @staticmethod
    def build_variables_array(name, sizes, domain, indexes=[]):
        if isinstance(name, list):
            # it means that several variables are declared with a single line
            assert len(sizes) == 1, "When using several declarations, only one-dimensional arrays are allowed."
            return [Variable.build_variable(var_name.strip(), domain, indexes) for var_name in name]
        if sizes:
            t = []
            for i in range(sizes[0]):
                indexes.append(i)
                t.append(Variable.build_variables_array(name, sizes[1:], domain, indexes))
                indexes.pop()
            return t
        var_name = name + "[" + "][".join(str(i) for i in indexes) + "]"
        return Variable.build_variable(var_name, domain, indexes)

    def __init__(self, name, dom):
        self.id = name
        self.dom = dom
        pos = self.id.find("[")
        if pos == -1:
            self.indexes = None
        else:
            self.prefix, self.suffix = self.id[:pos], self.id[pos:]
            self.indexes = [int(v) for v in re.split("\]\[", self.suffix[1:-1])]

    def __invert__(self):
        return NotVariable(self)

    def __neg__(self):
        return NegVariable(self)

    def __hash__(self, *args, **kwargs):
        return object.__hash__(self, *args, **kwargs)

    def __repr__(self):
        return self.id


class VariableInteger(Variable):
    def __init__(self, name, dom):
        super().__init__(name, dom)


class VariableSymbolic(Variable):
    def __init__(self, name, dom):
        super().__init__(name, dom)



from pycsp3.problems.data.parsing import *

next_line()
data["height"] = next_int()
data["width"] = next_int()

next_line()
nBugTypes = next_int()

next_line()
bugTypesLength = [next_int() for i in range(nBugTypes)];

next_line()
nBugs = next_int()

next_line()
data["bugs"] = []
data["bugTypes"] = []
lst = [False] * nBugs
for index in range(nBugs):
    bug = {"row": next_int(), "col": next_int(), "type": next_int()}
    data["bugs"].append(bug)
    if lst[bug["type"]] is False: lst[bug["type"]] = []
    lst[bug["type"]].append(index)

for index, length in enumerate(bugTypesLength):
    data["bugTypes"].append({"length": length, "cells": lst[index]})



from pycsp3.problems.data.parsing import *

data["n"] = number_in(line())
data["prizes"] = [numbers_in(line)[1:] if i > 0 else numbers_in(line[line.index("["):])[1:] for i, line in enumerate(remaining_lines(skip_curr=True))]



"""
Problem 031 on CSPLib

Example of Execution:
  python3 Rack2.py -data=Rack_r2b.json
"""

from pycsp3 import *

nRacks, models, cardTypes = data
models.append(models[0].__class__(0, 0, 0))  # we add first a dummy model (0,0,0) ; we get the class of the used named tuples to build a new one
powers, sizes, costs = zip(*models)
cardPowers, cardDemands = zip(*cardTypes)
nModels, nTypes = len(models), len(cardTypes)

table = {(i, powers[i], sizes[i], costs[i]) for i in range(nModels)}

# m[i] is the model used for the ith rack
m = VarArray(size=nRacks, dom=range(nModels))

# p[i] is the power of the model used for the ith rack
p = VarArray(size=nRacks, dom=powers)

# s[i] is the size (number of connectors) of the model used for the ith rack
s = VarArray(size=nRacks, dom=sizes)

# c[i] is the cost (price) of the model used for the ith rack
c = VarArray(size=nRacks, dom=costs)

# nc[i][j] is the number of cards of type j put in the ith rack
nc = VarArray(size=[nRacks, nTypes], dom=lambda i, j: range(min(max(sizes), cardDemands[j]) + 1))

satisfy(
    # linking rack models with powers, sizes and costs
    [(m[i], p[i], s[i], c[i]) in table for i in range(nRacks)],

    # connector-capacity constraints
    [Sum(nc[i]) <= s[i] for i in range(nRacks)],

    # power-capacity constraints
    [nc[i] * cardPowers <= p[i] for i in range(nRacks)],

    # demand constraints
    [Sum(nc[:, j]) == cardDemands[j] for j in range(nTypes)],

    # tag(symmetry-breaking)
    [Decreasing(m), imply(m[0] == m[1], nc[0][0] >= nc[1][0])]
)

minimize(
    # minimizing the total cost being paid for all racks
    Sum(c)
)


# Note that:

# a) the introduced dummy model is not saved when using the option -dataexport because here we don't modify data.rackModels

# models = [{'power': 0, 'nConnectors': 0, 'price': 0}] + models
# [model.power for model in models], [model.nConnectors for model in models], [model.price for model in models]
# [cardType.power for cardType in cardTypes], [cardType.demand for cardType in cardTypes]



from pycsp3 import *

x = VarArray(size=10, dom=range(150))
y = Var(dom=range(150))

satisfy(
    [x[p] in range(100) for p in range(10)],
    [x[p] in {1, 2, 3} for p in range(10)],
    [x[p] not in range(100) for p in range(10)],
    [x[p] not in {1, 2, 3} for p in range(10)],
    [x[p] <= 5 for p in range(10)],
    [x[p] < 5 for p in range(10)],
    [x[p] >= 5 for p in range(10)],
    [x[p] > 5 for p in range(10)],
    [x[p] == 5 for p in range(10)],
    [x[p] != 5 for p in range(10)],

    y in range(50),
    y not in range(50),
    y in {1, 2, 3, 3},
    y not in {1, 2, 3},
    y <= 5,
    y < 5,
    y >= 5,
    y > 5,
    y == 5,
    y != 5
)



import os

from setuptools import setup, find_packages

__version__ = open(os.path.join(os.path.dirname(__file__), 'pycsp3/version.txt'), encoding='utf-8').read()

print("setup version", __version__)

setup(name='pycsp3',
      version=__version__,
      python_requires='>=3,!=3.8.*,!=3.9.*',
      author='Lecoutre Christophe, Szczepanski Nicolas',
      author_email='lecoutre@cril.fr, szczepanski@cril.fr',
      maintainer='Szczepanski Nicolas',
      maintainer_email='szczepanski@cril.fr',
      keywords='IA CP constraint modeling',
      classifiers=['Topic :: Scientific/Engineering :: Artificial Intelligence', 'Topic :: Education'],
      packages=find_packages(exclude=["problems/g7_todo/"]),
      package_dir={'pycsp3': 'pycsp3'},
      install_requires=['lxml', 'py4j', 'numpy'],
      include_package_data=True,
      description='Modeling Constrained Combinatorial Problems in Python',
      long_description=open(os.path.join(os.path.dirname(__file__), 'pycsp3/README.md'), encoding='utf-8').read(),
      long_description_content_type='text/markdown',
      license='MIT',
      platforms='LINUX')



from pycsp3 import *

n = 5
x = VarArray(size=[n, n], dom=range(1, n + 1))

table = list(permutations(range(1, n + 1)))

try:
    x in table
except AssertionError as e:
    print("Scope and tuples not compatible:", e)

satisfy(
    # supports
    x[0] in table,

    # conflicts
    x[0] not in table
)
try:
    satisfy([1] * len(x[0]) in table)
except TypeError as e:
    print("Parameter 1:", e)

try:
    satisfy(x[0] in None)
except TypeError as e:
    print("Parameter 1:", e)

table[1] = (1, 2)
satisfy(
    # Wrong but too costly to check all lengths
    x[0] in table
)

table = list(permutations(range(1, n + 1)))
table[1] = ("a", "b", "c", "d", "e")

try:
    satisfy(x[0] in table)
except TypeError as e:
    print("Table with different types:", e)

table = [(1, 2, 3, 4, 5), (0.1, 0.2, 0.3, 0.4, 0.5)]
satisfy(
    #  float and int in table
    x[0] in table
)

satisfy(
    # supports
    x[0] in table,

    # conflicts
    x[0] not in table
)

try:
    satisfy(x[0][0] in table)
except AssertionError as e:
    print("Scope and tuples not compatible:", e)

try:
    satisfy(x[0] in table[0])
except AssertionError as e:
    print("Bad type for operand of in:", e)

# TODO
satisfy(
    # TODO is wrong or not ?
    x[0][0] in [0, 1]
)



"""
See https://en.wikipedia.org/wiki/Boolean_satisfiability_problem

Example of Execution:
  python3 Sat.py -data=Sat_flat30-16.json -variant=clause
  python3 Sat.py -data=Sat_flat30-16.json -variant=sum
  python3 Sat.py -data=Sat_flat30-16.json -variant=dual
"""

from pycsp3 import *

n, e, clauses = data


def scope(clause):
    return [x[abs(j) - 1] for j in clause]


def phases(clause):
    return [j >= 0 for j in clause]


if variant("clause"):
    # x[i] is the ith propositional variable
    x = VarArray(size=n, dom={0, 1})

    satisfy(
        Clause(scope(clause), phases=phases(clause)) for clause in clauses
    )

elif variant("sum"):
    # x[i] is the ith propositional variable
    x = VarArray(size=n, dom={0, 1})

    satisfy(
        scope(clause) * [1 if j >= 0 else -1 for j in clause] != -len([j for j in clause if j < 0]) for clause in clauses
    )

elif variant("dual"):  # dual construction [Bacchus, Extending forward checking, 2000]
    def dual_table(i, j):
        def base_value(decimal_value, length, base):
            value = []
            for _ in range(length):
                value = [decimal_value % base] + value
                decimal_value = decimal_value // base
            assert decimal_value == 0, "The given array is too small to contain all the digits of the conversion"
            return value

        def atom_value_at(clause, phasedLitPos, value):
            pos = phasedLitPos if phasedLitPos >= 0 else -phasedLitPos - 1
            return base_value(value, len(clause), 2)[pos] == (1 if phasedLitPos >= 0 else 0)  # > 0 = positive atom

        def check(clause1, clause2, a, b, links):
            return all(atom_value_at(clause1, link[0], a) == atom_value_at(clause2, link[1], b) for link in links)

        c1, c2 = clauses[i], clauses[j]
        links = [(i if c1[i] > 0 else -i - 1, j if c2[j] > 0 else -j - 1) for i in range(len(c1)) for j in range(len(c2)) if abs(c1[i]) == abs(c2[j])]
        return None if len(links) == 0 else {(v1, v2) for v1 in range(1, 2 ** len(c1)) for v2 in range(1, 2 ** len(c2)) if check(c1, c2, v1, v2, links)}


    x = VarArray(size=e, dom=lambda i: range(1, 2 ** len(clauses[i])))

    satisfy(
        (x[i], x[j]) in dual_table(i, j) for i, j in combinations(range(e), 2) if dual_table(i, j)
    )



from pandas import *
from pycsp3.problems.data.parsing import *


def createObject(name, file):
    xls = ExcelFile(file)
    df = xls.parse(xls.sheet_names[0])
    data[name] = dict()
    for key, value in df.to_dict().items():
        data[name][key] = list(value.values())

createObject('capacites', 'pycsp3/problems/data/raw/Planeparking/capacites.xlsx')
createObject('ombrages', 'pycsp3/problems/data/raw/Planeparking/ombrages.xlsx')
createObject('ordonnancement', 'pycsp3/problems/data/raw/Planeparking/ordonnancement.xlsx')
createObject('parkings', 'pycsp3/problems/data/raw/Planeparking/parkings.xlsx')
createObject('reductions', 'pycsp3/problems/data/raw/Planeparking/reductions.xlsx')
createObject('strategies', 'pycsp3/problems/data/raw/Planeparking/strategies_matrice.xlsx')
createObject('ordonnancement', 'pycsp3/problems/data/raw/Planeparking/ordonnancement.xlsx')
createObject('traitement', 'pycsp3/problems/data/raw/Planeparking/temps_traitement.xlsx')
createObject('vols', 'pycsp3/problems/data/raw/Planeparking/vols.xlsx')
createObject('volsStrategies', 'pycsp3/problems/data/raw/Planeparking/vols_strat.xlsx')





from pycsp3 import *

n = data
k, m = n, n  # k : Number of times round the loop ; m : m^2 propagators per change of loop ; n : Number of iterations of change per loop

x = VarArray(size=m + 1, dom=range(k * n + 1))

y = VarArray(size=n + 1, dom=range(k * n + 1))

satisfy(
    [y[i - 1] - y[i] <= 0 for i in range(2, n + 1)],

    [y[0] - y[i] <= n - i + 1 for i in range(1, n + 1)],

    y[n] - x[0] <= 0,

    [x[i] - x[j] <= 0 for i in range(m + 1) for j in range(i + 1, m + 1)],

    x[m] - y[0] <= -2
)



"""
For a given (undirected) graph G, the problem consists in arranging the nodes of the graph in a line
in such a way to minimize the sum of distances between adjacent nodes (in G).

Example of Execution:
  python3 LinearArrangement.py -data=MinLA01.json
"""

from pycsp3 import *

n, edges = data
edges = [(i, j) for i, j in edges]  # to be sure to have tuples

# x[i] denotes the position (in the line) of the ith node
x = VarArray(size=n, dom=range(n))

# d[i][j] denotes the distance in the line between the ith and jth nodes (if they are adjacent in the graph)
d = VarArray(size=[n, n], dom=lambda i, j: range(1, n) if (i, j) in edges else None)

satisfy(
    # putting nodes at different positions
    AllDifferent(x),

    # computing distances
    [d[i][j] == abs(x[i] - x[j]) for (i, j) in edges],

    # triangle constraints: distance(i,j) <= distance(i,k) + distance(k,j)  tag(redundant-constraints)
    [d[i][j] <= d[min(i, k)][max(i, k)] + d[min(j, k)][max(j, k)] for (i, j) in edges
     for k in range(n) if (min(i, k), max(i, k)) in edges and (min(j, k), max(j, k)) in edges]
)

minimize(
    # minimizing the sum of distances between adjacent nodes
    Sum(d)
)


# Note that for large instances, one may think about using an adjacency matrix, as with:
# a = [[1 if (i, j) in e or (j, i) in e else 0 for j in range(n)] for i in range(n)]



from pycsp3 import *

# x[i] is the ith value of the sequence
x = VarArray(size=4, dom=range(15))

satisfy(
    # four successive values are needed
    [x[i] + 1 == x[i + 1] for i in range(3)],

    # values must sum up to 14
    Sum(x) == 14
)



from pycsp3.problems.data.parsing import *


def linear_objective():
    tokens = line().split()
    assert len(tokens) % 2 == 0 and tokens[0] == "min:" and tokens[- 1] == ";"
    coeffs, nums = [int(tokens[i]) for i in range(1, len(tokens) - 1, 2)], [int(tokens[i][1:]) - 1 for i in range(2, len(tokens) - 1, 2)]
    return OrderedDict([("coeffs", coeffs), ("nums", nums)])


def linear_constraint():
    tokens = line().split()
    assert len(tokens) % 2 == 1 and tokens[- 1] == ";"
    coeffs, nums = [int(tokens[i]) for i in range(0, len(tokens) - 3, 2)], [int(tokens[i][1:]) - 1 for i in range(1, len(tokens) - 3, 2)]
    return OrderedDict([("coeffs", coeffs), ("nums", nums), ("op", tokens[- 3]), ("limit", int(tokens[- 2]))])


t = numbers_in(line())
data["n"] = t[0]
data["e"] = t[1]

while next_line()[0] == "*":  # comments
    pass

data["constraints"] = []
if line()[0] == "m":
    data["objective"] = linear_objective()
else:
    data["objective"] = None
    data["constraints"].append(linear_constraint())

while next_line() is not None:
    if line()[0] != "*":
        data["constraints"].append(linear_constraint())

assert len(data["constraints"]) == data["e"]



"""
Problem 031 on CSPLib

Example of Execution:
  python3 Rack.py -data=Rack_r2.json
"""

from pycsp3 import *

nRacks, models, cardTypes = data
models.append([0, 0, 0])  # we add first a dummy model (0,0,0)
powers, sizes, costs = zip(*models)
cardPowers, cardDemands = zip(*cardTypes)
nModels, nTypes = len(models), len(cardTypes)

table = {(i, powers[i], sizes[i], costs[i]) for i in range(nModels)}

# m[i] is the model used for the ith rack
m = VarArray(size=nRacks, dom=range(nModels))

# p[i] is the power of the model used for the ith rack
p = VarArray(size=nRacks, dom=powers)

# s[i] is the size (number of connectors) of the model used for the ith rack
s = VarArray(size=nRacks, dom=sizes)

# c[i] is the cost (price) of the model used for the ith rack
c = VarArray(size=nRacks, dom=costs)

# nc[i][j] is the number of cards of type j put in the ith rack
nc = VarArray(size=[nRacks, nTypes], dom=lambda i, j: range(min(max(sizes), cardDemands[j]) + 1))

satisfy(
    # linking rack models with powers, sizes and costs
    [(m[i], p[i], s[i], c[i]) in table for i in range(nRacks)],

    # connector-capacity constraints
    [Sum(nc[i]) <= s[i] for i in range(nRacks)],

    # power-capacity constraints
    [nc[i] * cardPowers <= p[i] for i in range(nRacks)],

    # demand constraints
    [Sum(nc[:, j]) == cardDemands[j] for j in range(nTypes)],

    # tag(symmetry-breaking)
    [Decreasing(m), imply(m[0] == m[1], nc[0][0] >= nc[1][0])]
)

minimize(
    # minimizing the total cost being paid for all racks
    Sum(c)
)

# Note that:

# a)  using zip is compacter than writing:
#   powers, sizes, costs = [row[0] for row in models], [row[1] for row in models], [row[2] for row in models]
#   cardPowers, cardDemands =[row[0] for row in cardTypes], [row[1] for row in cardTypes]

# b) using a quaternary table constraint is simpler than using three binary table constraints, as below:
#   linking model and power of the ith rack
#   [(m[i], p[i]) in enumerate(powers) for i in range(nRacks)],
#   linking model and size of the ith rack
#   [(m[i], s[i]) in enumerate(sizes) for i in range(nRacks)],
#   linking model and cost of the ith rack
#   [(m[i], c[i]) in enumerate(costs) for i in range(nRacks)],



"""
Problem 045 on CSPLib

It is possible to get the covering array from v (the array of variables in the model below).
For example, v[0][0] gives the t most significant bits of the first column (because the first t-combination is for the first t lines).

Examples of Execution:
  python3 CoveringArray.py
  python3 CoveringArray.py -data=[3,5,2,10]
"""

from pycsp3 import *
from math import factorial

t, k, g, b = data or (3, 5, 2, 10)
n = factorial(k) // factorial(t) // factorial(k - t)
d = g ** t

table = {tuple(sum(pr[a] * g ** i for i, a in enumerate(reversed(co))) for co in combinations(range(k), t)) for pr in product(range(g), repeat=k)}

# p[i][j] is one of the position of the jth value of the ith 't'-combination
p = VarArray(size=[n, d], dom=range(b))

# v[i][j] is the jth value of the ith 't'-combination
v = VarArray(size=[n, b], dom=range(d))

satisfy(
    # all values must be present in each 't'-combination
    [AllDifferent(p[i]) for i in range(n)],

    [Channel(p[i], v[i]) for i in range(n)],

    # computing values
    [v[:, j] in table for j in range(b)]
)



"""
See "Teaching Constraints through Logic Puzzles" by Peter Szeredi

"""

from pycsp3 import *

n, m = 7, 4  # data.n, data.m
# assert 0 < m <= n
# clues = data.clues
nSegments = ((n - 3) // 2) * 4  # it is correct for n = 7, but for other orders?

value_sequences = [tuple(1 + (i + j) % m for j in range(m)) for i in range(m)]


def scope(k):
    assert k > 0
    if k == 1:
        return x[0]
    if k == 2:
        return x[:, -1]
    if k == 3:
        return list(reversed(x[-1]))
    gap = k // 4 - 1
    if k % 4 == 0:
        return list(reversed(x[gap + 1:n if gap == 0 else - gap, gap]))  # top
    if k % 4 == 1:
        return x[gap + 1, gap:-1 - gap]  # right
    if k % 4 == 2:
        return x[gap + 1: -1 - gap, -2 - gap]  # down
    assert k % 4 == 3
    return list(reversed(x[-2 - gap, gap + 1:-1 - gap]))  # left


# x[i][j] is the value at row i and column j
x = VarArray(size=[n, n], dom=range(m + 1))

segment_scopes = cp_array(scope(k + 1) for k in range(nSegments))  # cp_array is required here for being able to use the constraint 'element'

satisfy(
    # putting each value on each row, and also the special value 0
    [Cardinality(x[i], occurrences={0: n - m} + {v: 1 for v in range(1, m + 1)}) for i in range(n)],

    # putting each value on each column, and also the special value 0
    [Cardinality(x[:, j], occurrences={0: n - m} + {v: 1 for v in range(1, m + 1)}) for j in range(n)],
)

if not variant():
    # p[k][l] is the position of the lth (non-zero) value of the kth segment
    p = VarArray(size=[nSegments, m], dom=lambda k, l: range(len(segment_scopes[k])))

    # v[k][l] is the lth (non-zero) value of the kth segment
    v = VarArray(size=[nSegments, m], dom=range(1, m + 1))

    satisfy(
        # positions must be ordered
        [Increasing(p[k], strict=True) for k in range(nSegments)],

        # the first segment has a fixed sequence of values
        v[0] == list(range(1, m + 1)),

        # other segments must respect authorized sequences of values
        [v[k] in value_sequences for k in range(1, nSegments)],

        # linking variables from x, p and v by means of constraints 'element'
        [segment_scopes[k][p[k][l]] == v[k][l] for k in range(nSegments) for l in range(m)]
    )


elif variant("table"):
    def table(k):
        r = len(segment_scopes[k])
        assert m <= r
        tbl = set()
        for combination in combinations(range(r), m):
            for sequence in [tuple(range(1, m + 1))] if k == 0 else value_sequences:
                t = [0] * r
                for i in range(m):
                    t[combination[i]] = sequence[i]
                tbl.add(tuple(t))
        return tbl


    satisfy(
        # enforcing each segment to contain values in ordered sequences
        [scp in table(k) for k, scp in enumerate(segment_scopes)]
    )



"""
 Problem 068 on CSPLib

Examples of Execution:
  python3 TravelingTournamentWithPredefinedVenues.py -data=Ttppv_circ8bbal.json -variant=a2
  python3 TravelingTournamentWithPredefinedVenues.py -data=Ttppv_circ8bbal.json -variant=a3
python3 pycsp3/problems/g4_world/TravelingTournamentWithPredefinedVenues.py -data=pycsp3/problems/data/json/Ttppv_circ8bbal.json -variant=a2
"""

from pycsp3 import *

nTeams, venues = data
nRounds = nTeams - 1
assert nTeams % 2 == 0, "an even number of teams is expected"
nConsecutiveGames = 2 if variant("a2") else 3  # used in one comment


def circ_distance(i, j):
    return min(abs(i - j), nTeams - abs(i - j))


def table_end(i):
    # when playing at home (whatever the opponent, travel distance is 0)
    return {(1, ANY, 0)} | {(0, j, circ_distance(i, j)) for j in range(nTeams) if j != i}


def table_intern(i):
    return ({(1, 1, ANY, ANY, 0)} |
            {(0, 1, j, ANY, circ_distance(j, i)) for j in range(nTeams) if j != i} |
            {(1, 0, ANY, j, circ_distance(i, j)) for j in range(nTeams) if j != i} |
            {(0, 0, j1, j2, circ_distance(j1, j2)) for j1 in range(nTeams) for j2 in range(nTeams) if different_values(i, j1, j2)})


def automaton():
    tr2 = [("q", 0, "q01"), ("q", 1, "q11"), ("q01", 0, "q02"), ("q01", 1, "q11"), ("q11", 0, "q01"), ("q11", 1, "q12"), ("q02", 1, "q11"), ("q12", 0, "q01")]
    tr3 = [("q02", 0, "q03"), ("q12", 1, "q13"), ("q03", 1, "q11"), ("q13", 0, "q01")]
    fi2 = {"q01", "q02", "q11", "q12"}
    fi3 = {"q03", "q13"}
    return Automaton(start="q", final=fi2 if variant("a2") else fi2 | fi3, transitions=tr2 if variant("a2") else tr2 + tr3)


# o[i][k] is the opponent (team) of the ith team  at the kth round
o = VarArray(size=[nTeams, nRounds], dom=range(nTeams))

# h[i][k] is 1 iff the ith team plays at home at the kth round
h = VarArray(size=[nTeams, nRounds], dom={0, 1})

# t[i][k] is the travelled distance by the ith team at the kth round. An additional round is considered for returning at home.
t = VarArray(size=[nTeams, nRounds + 1], dom=range(nTeams // 2 + 1))

satisfy(
    # a team cannot play against itself
    [o[i][k] != i for i in range(nTeams) for k in range(nRounds)],

    # ensuring predefined venues  
    [venues[i][o[i][k]] == h[i][k] for i in range(nTeams) for k in range(nRounds)],

    # ensuring symmetry of games: if team i plays against j, then team j plays against i
    [o[:, k][o[i][k]] == i for i in range(nTeams) for k in range(nRounds)],

    # each team plays once against all other teams
    [AllDifferent(row) for row in o],

    # at most 'nConsecutiveGames' consecutive games at home, or consecutive games away
    [h[i] in automaton() for i in range(nTeams)],

    # handling travelling for the first game
    [(h[i][0], o[i][0], t[i][0]) in table_end(i) for i in range(nTeams)],

    # handling travelling for the last game
    [(h[i][- 1], o[i][- 1], t[i][-1]) in table_end(i) for i in range(nTeams)],

    # handling travelling for two successive games
    [(h[i][k], h[i][k + 1], o[i][k], o[i][k + 1], t[i][k + 1]) in table_intern(i) for i in range(nTeams) for k in range(nRounds - 1)],

    # at each round, opponents are all different  tag(redundant-constraints)
    [AllDifferent(col) for col in columns(o)],

    # tag(symmetry-breaking)
    o[0][0] < o[0][- 1]
)

minimize(
    # minimizing summed up travelled distance
    Sum(t)
)



"""
See "Teaching Constraints through Logic Puzzles" by Peter Szeredi

Example of Execution:
  python3 Auction.py -data=Areas-3-3-3.json
"""

from pycsp3 import *

puzzle = data
n, m = len(puzzle), len(puzzle[0])

Region = namedtuple("Region", "i j size")
regions = [Region(i + 1, j + 1, puzzle[i][j]) for i in range(n) for j in range(m) if puzzle[i][j] != 0]  # +1 because a border is inserted
nRegions = len(regions)

conflicting_regions = {(k1, k2) for k1 in range(nRegions) for k2 in range(nRegions) if k1 != k2 and regions[k1].size == regions[k2].size}


def table_start(k):  # to be used with the starting square (where clue is given) of region k
    return {(k, k, ANY, ANY, ANY, 0, 1, ANY, ANY, ANY),
            (k, ANY, k, ANY, ANY, 0, ANY, 1, ANY, ANY),
            (k, ANY, ANY, k, ANY, 0, ANY, ANY, 1, ANY),
            (k, ANY, ANY, ANY, k, 0, ANY, ANY, ANY, 1)}


def table_other():
    t = set()
    for k in range(nRegions):
        for v in range(1, regions[k].size):
            t.update([
                (k, k, ANY, ANY, ANY, v, v - 1, ANY, ANY, ANY),
                (k, ANY, k, ANY, ANY, v, ANY, v - 1, ANY, ANY),
                (k, ANY, ANY, k, ANY, v, ANY, ANY, v - 1, ANY),
                (k, ANY, ANY, ANY, k, v, ANY, ANY, ANY, v - 1)
            ])
    return t


def scope(i, j):
    return x[i][j], x[i][j - 1], x[i][j + 1], x[i - 1][j], x[i + 1][j], d[i][j], d[i][j - 1], d[i][j + 1], d[i - 1][j], d[i + 1][j]


# x[i][j] is the region (number) where the square at row i and column j belongs (borders are inserted for simplicity)
x = VarArray(size=[n + 2, m + 2], dom=lambda i, j: {-1} if i in {0, n + 1} or j in {0, m + 1} else range(nRegions))

# d[i][j] is the distance of the square at row i and column j wrt the starting square of the (same) region
d = VarArray(size=[n + 2, m + 2],
             dom=lambda i, j: {-1} if i in {0, n + 1} or j in {0, m + 1} else {0} if puzzle[i - 1][j - 1] != 0 else range(1,
                                                                                                                          max(r.size for r in regions)))

satisfy(
    # setting starting squares of regions
    [x[i][j] == k for k, (i, j, _) in enumerate(regions)],

    # respecting the size of each region
    [Count(x, value=k) == s for k, (_, _, s) in enumerate(regions)],

    # two regions of the same size cannot have neighbouring squares
    [
        [(x[i][j], x[i][j + 1]) not in conflicting_regions for i in range(1, n + 1) for j in range(1, m)],
        [(x[i][j], x[i + 1][j]) not in conflicting_regions for j in range(1, m + 1) for i in range(1, n)]
    ],

    # each starting square of a (non-unit) region must have at least one neighbor at distance 1
    [scope(i, j) in table_start(k) for k, (i, j, s) in enumerate(regions) if s > 1],

    # each square must be connected to a neighbour at distance 1
    [scope(i, j) in table_other() for i in range(1, n + 1) for j in range(1, m + 1) if puzzle[i - 1][j - 1] == 0]
)



from pycsp3 import *

order, nQueens, nPions= data

q = VarArray(size=nQueens, dom=range(order))

p = VarArray(size=nPions, dom=range(order * order))

satisfy(
    [(q[i] != q[j]) & (dist(q[i], q[j]) != dist(i, j)) for i in range(nQueens) for j in range(i + 1, nQueens)],

    [(dist(p[i], p[j]) < nPions - 1) & (p[i] != p[j]) for i in range(nPions) for j in range(i + 1, nPions)]
)

if variant("mul"):
    satisfy(
        (q[i] != p[j] % order) | (i != p[j] // order) for i in range(nQueens) for j in range(nPions)
    )


"""
A toy problem from my course about CP.

Execution:
  python3 Witch.py
"""

from pycsp3 import *

# x is the number of magic potions for love
x = Var(range(400))

# y is the number of magic potions for youth
y = Var(range(400))

satisfy(
    [x, y] * [3, 2] <= 800,
    [x, y] * [1, 3] <= 700,
    [x, y] * [1, 2] <= 400
)

maximize(
    4 * x + 5 * y
)



class _Options:
    def __init__(self):
        self.values = tuple()  # options with non-Boolean values (strings or numbers)
        self.flags = tuple()  # Boolean options
        self.parameters = []
        self.parameters_cursor = 0

    def set_values(self, *values):
        self.values = [value.lower() for value in values]
        for option in self.values:
            vars(self)[option] = None

    def set_flags(self, *flags):
        self.flags = [flag.lower() for flag in flags]
        for option in self.flags:
            vars(self)[option] = False

    def get(self, name):
        return vars(self)[name]

    def consume_parameter(self):
        if self.parameters_cursor < len(self.parameters):
            parameter = self.parameters[self.parameters_cursor]
            self.parameters_cursor += 1
            return parameter
        else:
            return None

    def parse(self, args):
        for arg in args:
            if arg[0] == '-':
                t = arg[1:].split('=', 1)
                if len(t) == 1:
                    flag = t[0].lower()
                    if flag in self.flags:
                        vars(self)[flag] = True
                        assert flag not in self.values or flag == 'dataexport', "You have to specify a value for the option -" + flag
                    else:
                        print("Warning: Unknown option", arg)
                else:
                    assert len(t) == 2
                    value = t[0].lower()
                    if value in self.values:
                        assert len(t[1]) > 0, "The value specified for the option -" + value + " is the empty string"
                        vars(self)[value] = t[1]
                    else:
                        print("Warning: Unknown option", arg)
            else:
                self.parameters.append(arg)


options = _Options()



from pycsp3.classes.auxiliary.ptypes import TypeCtr, TypeCtrArg
from pycsp3.classes.main.constraints import ConstraintUnmergeable


class Objective(ConstraintUnmergeable):
    def __init__(self, way):
        assert way in {TypeCtr.MINIMIZE, TypeCtr.MAXIMIZE}
        super().__init__(way)


class ObjectiveExpression(Objective):
    def __init__(self, way, expression):
        super().__init__(way)
        self.arg(TypeCtrArg.EXPRESSION, expression)


class ObjectivePartial(Objective):
    def __init__(self, way, partial):
        super().__init__(way)
        self.attributes.append((TypeCtrArg.TYPE, partial.constraint.name))
        self.arguments = partial.constraint.arguments



from pycsp3 import *

n = data  # number of pigeons

# p[i] is the hole where is put the ith pigeon
p = VarArray(size=n, dom=range(n - 1))

if not variant():
    satisfy(
        AllDifferent(p)
    )

elif variant("dec"):
    satisfy(
        p[i] != p[j] for i in range(n) for j in range(i + 1, n)
    )



"""
The Multi dimensional knapsack problem was originally proposed as an optimization problem by the OR community.
Here, it is the feasability version, as used, e.g., in (Refalo, CP 2004) and (Pesant et al., JAIR 2012.

Example of Execution:
  python3 MultiKnapsack.py -data=MultiKnapsack_example.json
"""

from pycsp3 import *

weights, constraints = data
n = len(weights)

# x[i] is 1 iff the ith item is selected
x = VarArray(size=n, dom={0, 1})

satisfy(
    x * coeffs <= k for (coeffs, k) in constraints
)

maximize(
    x * weights
)



class SymbolicValue:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value if isinstance(other, SymbolicValue) else False

    def __lt__(self, other):
        assert isinstance(other, SymbolicValue)
        return self.value < other.value

    def __contains__(self, v):
        return v == self.value

    def __repr__(self):
        return str(self.value)


class IntegerEntity:
    def smallest(self):
        pass

    def greatest(self):
        pass

    def width(self):
        pass


class IntegerValue(IntegerEntity):
    def __init__(self, value):
        self.value = value

    def smallest(self):
        return self.value

    def greatest(self):
        return self.value

    def width(self):
        return 1

    def to_list(self):
        return [self.value]

    def __eq__(self, other):
        return self.value == other.value if isinstance(other, IntegerValue) else False

    def __lt__(self, other):
        assert isinstance(other, IntegerValue)
        return self.value < other.value

    def __contains__(self, v):
        return v == self.value

    def __repr__(self):
        return str(self.value)


class IntegerInterval(IntegerEntity):
    def __init__(self, inf, sup):
        assert inf < sup
        self.inf = inf
        self.sup = sup
        self.width = float('inf') if inf == float('-inf') or sup == float('inf') else sup + 1 - inf

    def smallest(self):
        return self.inf

    def greatest(self):
        return self.sup

    def width(self):
        return self.width

    def to_list(self):
        return [v for v in range(self.inf, self.sup + 1)]

    def is_binary(self):
        return self.inf == 0 and self.sup == 1

    def __iter__(self):
        return self.to_list().__iter__()

    def __getitem__(self, item):
        return self.to_list().__getitem__(item)

    def __eq__(self, other):
        return self.inf == other.inf and self.sup == other.sup if isinstance(other, IntegerInterval) else False

    def __lt__(self, other):
        assert isinstance(other, IntegerInterval)
        return self.sup <= other.inf

    def __contains__(self, v):
        return self.inf <= v <= self.sup

    def __repr__(self):
        return str(self.inf) + ".." + str(self.sup)



from pycsp3.problems.data.parsing import *

# python3 pycsp3/problems/g3_pattern/HCPizza.py -data=pycsp3/problems/data/raw/HCPizza_tiny.txt -dataparser=pycsp3/problems/data/parsers/HCPizza_Parser.py -variant=sum

nRows = next_int()
next_int()  # nCols
data["minIngredients"] = next_int()
data["maxSize"] = next_int()
data["pizza"] = [[0 if c == "M" else 1 for c in line if c != '\n'] for line in remaining_lines()]



"""
See, e.g., "Constraint programming lessons learned from crossword puzzles" by A. Beacham, X. Chen, J. Sillito, and P. van Beek. 2001

Examples of Execution:
  python3 Crossword.py -data=Crossword_vg6-7-ogd.json
  python3 Crossword.py -data=Crossword_vg6-7-ogd.json -variant=alt
"""

from pycsp3 import *

spots, dict_name = data
words = dict()
for line in open(dict_name):
    code = alphabet_positions(line.strip().lower())
    words.setdefault(len(code), []).append(code)


def find_holes(matrix, transposed):
    def build_hole(row, col, size, horizontal):
        return Hole(row, slice(col, col + size), size) if horizontal else Hole(slice(col, col + size), row, size)

    Hole = namedtuple("Hole", "i j r")  # i and j are indexes (one of them being a slice) and r is the size
    p, q = len(matrix), len(matrix[0])
    t = []
    for i in range(p):
        start = -1
        for j in range(q):
            if matrix[i][j] == 1:
                if start != -1 and j - start >= 2:
                    t.append(build_hole(i, start, j - start, not transposed))
                start = -1
            elif start == -1:
                start = j
            elif j == q - 1 and q - start >= 2:
                t.append(build_hole(i, start, q - start, not transposed))
    return t


holes = find_holes(spots, False) + find_holes(transpose(spots), True)
arities = sorted(set(size for (_, _, size) in holes))
n, m, nHoles = len(spots), len(spots[0]), len(holes)

if not variant():
    #  x[i][j] is the letter, number from 0 to 25, at row i and column j (when no spot)
    x = VarArray(size=[n, m], dom=lambda i, j: range(26) if spots[i][j] == 0 else None)

    satisfy(
        # fill the grid with words
        [x[i, j] in words[r] for (i, j, r) in holes],

        # tag(distinct-words)
        [AllDifferentList(x[i, j] for (i, j, r) in holes if r == arity) for arity in arities]
    )

elif variant("alt"):
    def offset(hole1, hole2):
        if type(hole1.i) == type(hole2.i):  # it means that they are both horizontal or vertical
            return None
        if isinstance(hole1.i, int):  # if hole1 is horizontal (and so hole2 is vertical)
            ofs1, ofs2 = hole2.j - hole1.j.start, hole1.i - hole2.i.start
            return (ofs1, ofs2) if 0 <= ofs1 < hole1.r and 0 <= ofs2 < hole2.r else None
        else:  # if hole1 is vertical (and so hole2 is horizontal)
            ofs1, ofs2 = hole1.j - hole2.j.start, hole2.i - hole1.i.start
            return (ofs1, ofs2) if 0 <= ofs1 < hole2.r and 0 <= ofs2 < hole1.r else None


    def table_compatible_words(hole1, hole2):
        ofs1, ofs2 = offset(hole1, hole2)
        return [(i1, i2) for i1, word1 in enumerate(words[hole1.r]) for i2, word2 in enumerate(words[hole2.r]) if word1[ofs1] == word2[ofs2]]


    # w[i] is the ith word to be put in the grid
    w = VarArray(size=nHoles, dom=lambda i: range(len(words[holes[i].r])))

    satisfy(
        # words must intersect correctly
        [(w[i], w[j]) in table_compatible_words(holes[i], holes[j]) for i, j in combinations(range(nHoles), 2) if offset(holes[i], holes[j])],

        # tag(distinct-words)
        [w[i] != w[j] for i, j in combinations(range(nHoles), 2) if holes[i].r == holes[j].r]
    )


# Note that:

# a) we use lists instead of sets for tables ([(i1, i2)  ... instead of  {(i1, i2) ..) because it is quite faster to process

# b) it is not possible to write x[i][j] when i is a slice; this must be x[i, j]



from pycsp3 import *

limit, value = data or (364, 55440)


def decompositions():
    t = []
    for i in range(1, limit + 1):
        if i * limit * (limit - 1) * (limit - 2) < value:
            continue
        if i * (i + 1) * (i + 2) * (i + 3) > value:
            break
        for j in range(i + 1, limit + 1):
            if i * j * limit * (limit - 1) < value:
                continue
            if i * j * (j + 1) * (j + 2) > value:
                break
            for k in range(j + 1, limit + 1):
                if i * j * k * limit < value:
                    continue
                if i * j * k * (k + 1) > value:
                    break
                for l in range(k + 1, limit + 1):
                    if i * j * k * l > value:
                        break
                    if i * j * k * l == value:
                        t.append((i, j, k, l))
    return t


table = {(t[i], t[j], t[k], t[l]) for t in decompositions() for (i, j, k, l) in product(range(4), repeat=4) if different_values(i, j, k, l)}

x = VarArray(size=4 * 4 * 4, dom=range(1, limit + 1))

satisfy(
    AllDifferent(x),
    [(x[i * 4], x[i * 4 + 1], x[i * 4 + 2], x[i * 4 + 3]) in table for i in range(16)],
    [([x[i], x[i + 4], x[i + 8], x[i + 12]]) in table for i in range(4)],
    [([x[i], x[i + 4], x[i + 8], x[i + 12]]) in table for i in range(16, 20)],
    [([x[i], x[i + 4], x[i + 8], x[i + 12]]) in table for i in range(32, 36)],
    [([x[i], x[i + 4], x[i + 8], x[i + 12]]) in table for i in range(48, 52)],
    [([x[i], x[i + 16], x[i + 32], x[i + 48]]) in table for i in range(16)],
    (x[0], x[21], x[42], x[63]) in table,
    (x[15], x[26], x[37], x[48]) in table,
    ([x[3], x[22], x[41], x[60]]) in table,
    (x[12], x[25], x[38], x[51]) in table
)



"""
See https://en.wikipedia.org/wiki/Strip_packing_problem

Example of Execution:
  python3 StripPacking.py -data=StripPacking_C1P1.json
"""

from pycsp3 import *

width, height = data.container
rectangles = data.rectangles
nRectangles = len(rectangles)

# x[i] is the x-coordinate of the ith rectangle
x = VarArray(size=nRectangles, dom=range(width))

# y[i] is the y-coordinate of the ith rectangle
y = VarArray(size=nRectangles, dom=range(height))

# w[i] is the width of the ith rectangle
w = VarArray(size=nRectangles, dom=lambda i: {rectangles[i].width, rectangles[i].height})

# h[i] is the height of the ith rectangle
h = VarArray(size=nRectangles, dom=lambda i: {rectangles[i].width, rectangles[i].height})

# r[i] is 1 iff the ith rectangle is rotated by 90 degrees
r = VarArray(size=nRectangles, dom={0, 1})

satisfy(
    # horizontal control
    [x[i] + w[i] <= width for i in range(nRectangles)],

    # vertical control
    [y[i] + h[i] <= height for i in range(nRectangles)],

    # managing rotation
    [(r[i], w[i], h[i]) in {(0, wgt, hgt), (1, hgt, wgt)} for i, (wgt, hgt) in enumerate(rectangles)],

    # no overlapping between rectangles
    NoOverlap(origins=[(x[i], y[i]) for i in range(nRectangles)], lengths=[(w[i], h[i]) for i in range(nRectangles)])
)



import re
from collections import OrderedDict

from pycsp3.classes.auxiliary.ptypes import TypeCtrArg
from pycsp3.classes.auxiliary.conditions import Condition
from pycsp3.classes.entities import CtrEntities, ECtr, ECtrs, EGroup, EBlock, EToGather, ESlide, EToSatisfy, TypeNode
from pycsp3.classes.main.constraints import ConstraintIntension, Diffs, ConstraintInstantiation

LIMIT_FOR_VAR_ARGS = 3


def build_similar_constraints():
    detecting_groups_recursively(CtrEntities.items)
    building_groups_recursively(CtrEntities.items)
    canonizing_groups_and_blocks(CtrEntities.items)
    recognizing_instantiations(CtrEntities.items)


# Phase 1: Detecting groups of similar constraints
def detecting_groups_recursively(ctr_entities):
    def _detecting_groups(entities):
        flags = [False] * len(entities)  # indicate (indexes of) constraints that are similar at a given moment (see intern loop)
        removal = False
        for i, e1 in enumerate(entities):
            if flags[i] or e1 is None or isinstance(e1, EGroup):
                continue
            Diffs.reset()
            group = EGroup()  # group of similar constraints tried to be built
            group.entities.append(e1)  # first constraint of the new group
            for j in range(i + 1, len(entities)):
                e2 = entities[j]
                if flags[j] is False and not isinstance(e2, EGroup):
                    diffs = e2.constraint.close_to(e1.constraint)
                    if diffs is not False:
                        diffs.merge()  # merging flags of two lists, indicating if the length of argument contents is different
                        group.entities.append(e2)  # adding the constraint to the new group
                        flags[j] = True  # to avoid processing it again
                        entities[j] = None  # to discard this constraint (since it is now in a group)
                        removal = True
            if len(group.entities) > 1:
                group.diff_argument_names = Diffs.fusion.argument_names
                group.diff_argument_flags = Diffs.fusion.argument_flags
                entities[i] = group  # constraint replaced by the new group (and other constraints of the group will be now ignored since flagged to False)

        return [e for e in entities if e is not None] if removal else entities

    for e in ctr_entities:
        if isinstance(e, (EBlock, ESlide, EToSatisfy)):
            detecting_groups_recursively(e.entities)
        if isinstance(e, EToGather):
            e.entities = _detecting_groups(e.entities)


# Phase 2: Building groups (constraint templates and args)


def _compute_group_abstraction_intension(group):
    def _replace_parameter(s, k, v):
        return (s.replace("(" + k + ")", "(" + v + ")").replace("," + k + ",", "," + v + ",").replace("(" + k + ",", "(" + v + ",")
                .replace("," + k + ")", "," + v + ")").replace("(" + k + ")", "(" + v + ")"))

    def _is_same_value_at_column(i, all_args):
        """ comparison from both ends so as to find opportunistically that the parameter is not the same everywhere"""
        left, right = 1, len(all_args) - 1
        value = all_args[0][i]
        while True:
            if value != all_args[left][i]:
                return False
            left += 1
            if value != all_args[right][i]:
                return False
            right -= 1
            if left > right:
                break
        return True

    def _is_same_value_in_columns(i, j, all_args):
        left, right = 0, len(all_args) - 1
        while True:
            if all_args[left][i] != all_args[left][j]:
                return False
            left += 1
            if all_args[right][i] != all_args[right][j]:
                return False
            right -= 1
            if left > right:
                break
        return True

    all_args = [ce.constraint.abstract_values() for ce in group.entities]
    abstract_tree = group.entities[0].constraint.abstract_tree()
    n_parameters = len(group.entities[0].constraint.abstract_values())
    marked = [False] * n_parameters

    # discarding a parameter if the same value appears all along the column
    for i in range(n_parameters):
        if _is_same_value_at_column(i, all_args):
            marked[i] = True
            abstract_tree = _replace_parameter(abstract_tree, "%" + str(i), str(all_args[0][i]))

    # merging two columns if they are identical
    for i in range(n_parameters):
        for j in range(i + 1, n_parameters):
            if _is_same_value_in_columns(i, j, all_args):
                marked[j] = True
                abstract_tree = _replace_parameter(abstract_tree, "%" + str(j), "%" + str(i))

    # removing marked arguments (in a symbolic way with a special value)
    for i in range(n_parameters - 1, -1, -1):
        if marked[i]:
            for args in all_args:
                del args[i]  # args[i] = DELETED

    # reordering %i in the abstract tree
    for i in range(n_parameters):
        par = "%" + str(i)
        if par not in re.findall('(%\d+)', abstract_tree):
            t = {int(m[1:]) for m in re.findall('(%\d+)', abstract_tree) if int(m[1:]) > i}
            if len(t) > 0:
                old = "%" + str(min(t))
                # while old in abstract_tree:
                abstract_tree = _replace_parameter(abstract_tree, old, par)
    return abstract_tree, all_args


def _compute_group_abstraction_other(group):
    c = group.entities[0].constraint
    abstraction = OrderedDict()
    var_args_argument = None

    # building the abstract constraint template
    abstract_condition = len(group.diff_argument_names) == 2 and group.diff_argument_names[-1] == TypeCtrArg.CONDITION
    for arg in c.arguments.values():
        if arg.name in group.diff_argument_names:
            index = group.diff_argument_names.index(arg.name)
            if group.diff_argument_flags[index]:
                assert len(set(len(ce.constraint.arguments[arg.name].content) for ce in group.entities)) > 1
                abstraction[arg.name] = "%..."
                var_args_argument = arg.name
            elif abstract_condition and index == 0 and isinstance(group.entities[0].constraint.arguments[arg.name].content, list) and len(
                    group.entities[0].constraint.arguments[arg.name].content) > LIMIT_FOR_VAR_ARGS:
                abstraction[arg.name] = "%..."
                var_args_argument = arg.name
            elif var_args_argument is None and group.diff_argument_names[- 1] == arg.name and isinstance(arg.content, list) and len(
                    arg.content) > LIMIT_FOR_VAR_ARGS:
                abstraction[arg.name] = "%..."
            else:
                if arg.name == TypeCtrArg.CONDITION:
                    abstraction[arg.name] = "(" + str(arg.content.operator) + "," + c.parameter_form(arg.content) + ")"  # TODO is it correct?
                else:
                    abstraction[arg.name] = c.parameter_form(arg.content)
        else:
            abstraction[arg.name] = arg.content

    def add_content(content):
        if isinstance(content, list):
            tmp.extend(content)
        elif isinstance(content, Condition):
            tmp.append(str(content.right_operand()))
        else:
            tmp.append(content)

    # collecting arguments
    all_args = []
    arguments_names = list(c.arguments.keys())
    for ce in group.entities:
        tmp = []
        if var_args_argument is None or group.diff_argument_flags[- 1]:
            for name in arguments_names:
                if name in group.diff_argument_names:
                    add_content(ce.constraint.arguments[name].content)
        else:
            for name in arguments_names:
                if name in group.diff_argument_names and name != var_args_argument:
                    add_content(ce.constraint.arguments[name].content)
            add_content(ce.constraint.arguments[var_args_argument].content)
        all_args.append(tmp)
    return abstraction, all_args


def building_groups_recursively(entities, previous=None):
    def _build_group(group):
        if isinstance(group.entities[0].constraint, ConstraintIntension):
            group.abstraction, group.all_args = _compute_group_abstraction_intension(group)
        else:
            group.abstraction, group.all_args = _compute_group_abstraction_other(group)

    for e in entities:
        if isinstance(e, EGroup):
            _build_group(e)
            e.copy_basic_attributes_of(previous)
            # previous.clearBasicAttributes()
        if isinstance(e, (ESlide, EToGather, EBlock, EToSatisfy)):
            building_groups_recursively(e.entities, e)


# Phase 3: adding/removing some blocks
def canonizing_groups_and_blocks(entities, previous=None):
    def _building_block(entity):
        if len(entity.entities) == 0:
            return entity
        if len(entity.entities) == 1:  # no need to create a block in this case
            entity.entities[0].copy_basic_attributes_of(entity)
            return entity.entities[0]
        block = EBlock([entity])  # creating a new block and copying the appropriate comments and tags
        block.copy_basic_attributes_of(entity if not entity.blank_basic_attributes() else entity.entities[0])
        for c in entity.entities:  # clearing comments and tags of the constraints embedded in the block
            c.clear_basic_attributes()
        return block

    for i, e in enumerate(entities):
        if not isinstance(e, ECtrs) or len(e.entities) == 0:
            continue
        first = e.entities[0]
        if isinstance(e, EBlock) and isinstance(first, (ECtr, EToGather, EGroup)):
            # removing a block when we have only one element inside
            if len(e.entities) == 1 and None in {e.comment, first.comment} and 0 in {len(e.tags), len(first.tags)}:
                if e.comment:
                    first.note(e.comment)
                if len(e.tags) > 0:
                    first.tag(e.tags)
                entities[i] = first
        if isinstance(e, EToGather):
            # Creating a new block when there are several sub-groups (or stand-alone constraints); it was impossible to do a single group
            if any(not c.blank_basic_attributes() for c in e.entities):
                if not isinstance(previous, EBlock):
                    entities[i] = _building_block(e)
            elif not e.blank_basic_attributes():
                entities[i] = _building_block(e)
        if isinstance(e, (ESlide, EBlock, EToSatisfy)):
            canonizing_groups_and_blocks(e.entities, e)


# Phase 4: detecting forms corresponding to instantiations
def recognizing_instantiations(entities):
    def _elements_for_building_instantiation(trees):
        t1, t2 = [], []
        for tree in trees:
            pair = tree.var_val_if_binary_type(TypeNode.EQ)
            if pair is None:
                return None
            t1.append(pair[0])
            t2.append(pair[1])
        return t1, t2

    for i, e in enumerate(entities):
        # since a group, we can test only the first constraint (to see if it is a group of intension constraints)
        if isinstance(e, EGroup) and isinstance(e.entities[0].constraint, ConstraintIntension):
            elements = _elements_for_building_instantiation(c.constraint.arguments[TypeCtrArg.FUNCTION].content for c in e.entities)
            if elements:
                entities[i] = ECtr(ConstraintInstantiation(elements[0], elements[1])).copy_basic_attributes_of(e)
        elif isinstance(e, (ESlide, EToGather, EBlock, EToSatisfy)):
            recognizing_instantiations(e.entities)



from types import GeneratorType

from pycsp3.classes.auxiliary.ptypes import TypeConditionOperator
from pycsp3.classes.main.variables import Variable
from pycsp3.tools.inspector import checkType


class Condition:
    def __init__(self, operator):
        self.operator = operator

    @staticmethod
    def build_condition(condition):
        if condition is None:
            return None  # it may occur when building a partial constraint
        condition = tuple(condition) if isinstance(condition, list) else condition  # we expect a condition to be given as a tuple (or a list)
        assert isinstance(condition, tuple) and len(condition) == 2, "a condition must a pair, given as a tuple (or a list)"
        operator = TypeConditionOperator.value_of(condition[0]) if isinstance(condition[0], str) else condition[0]
        assert isinstance(operator, TypeConditionOperator), "the operator " + str(operator) + " is not correct (should be of type TypeConditionOperator)"
        right_operand = list(condition[1]) if isinstance(condition[1], (set, frozenset, GeneratorType)) else condition[1]
        checkType(right_operand, (int, Variable, range, [int, Variable]))
        if isinstance(right_operand, range) and right_operand.step != 1:
            right_operand = list(right_operand)
        if isinstance(right_operand, int):
            return ConditionValue(operator, right_operand)
        if isinstance(right_operand, Variable):
            return ConditionVariable(operator, right_operand)
        if isinstance(right_operand, range):
            return ConditionInterval(operator, right_operand.start, right_operand.stop - 1)
        if isinstance(right_operand, list):
            return ConditionSet(operator, right_operand)

    def right_operand(self):
        pass

    def __str__(self):
        return "(" + str(self.operator) + "," + str(self.right_operand()) + ")"


class ConditionValue(Condition):
    def __init__(self, operator, value):
        super().__init__(operator)
        self.value = value

    def right_operand(self):
        return self.value

    def filtering(self, values):
        if self.operator == TypeConditionOperator.NE:
            return (v for v in values if v != self.value)
        if self.operator == TypeConditionOperator.LT:
            return (v for v in values if v < self.value)
        if self.operator == TypeConditionOperator.LE:
            return (v for v in values if v <= self.value)
        if self.operator == TypeConditionOperator.GE:
            return (v for v in values if v >= self.value)
        if self.operator == TypeConditionOperator.GT:
            return (v for v in values if v > self.value)
        assert False


class ConditionVariable(Condition):
    def __init__(self, operator, variable):
        super().__init__(operator)
        self.variable = variable

    def right_operand(self):
        return self.variable


class ConditionInterval(Condition):
    def __init__(self, operator, min, max):
        super().__init__(operator)
        self.min = min
        self.max = max

    def right_operand(self):
        return str(self.min) + ".." + str(self.max)


class ConditionSet(Condition):
    def __init__(self, operator, t):
        super().__init__(operator)
        self.t = t

    def right_operand(self):
        return "{" + ",".join(str(v) for v in self.t) + "}"


def ne(v):
    return ConditionValue(TypeConditionOperator.NE, v)



from pycsp3.problems.data.parsing import *

nRows, nCols = numbers_in(line())
next_line()

data["grid"] = [[5 if token == "a" else -1 if token == "-" else int(token) for token in [next_str() for col in range(nCols)]] for row in range(nRows)]



import os
import subprocess
import time
from io import IOBase

from lxml import etree
from py4j.java_gateway import JavaGateway, Py4JNetworkError

from pycsp3.classes.main.variables import Variable, VariableInteger
from pycsp3.classes.entities import VarEntities, EVarArray, EVar
from pycsp3.tools.utilities import Stopwatch, flatten
from pycsp3.dashboard import options


def directory_of_solver(name):
    # assert name == "abscon", "for the moment, only one embedded solver (abscon)"
    return os.sep.join(__file__.split(os.sep)[:-1]) + os.sep + name + os.sep


def class_path_abscon():
    d = directory_of_solver("abscon")
    return d + "AbsCon-20-01.jar" + os.pathsep + d + "xcsp3-tools-1.1.1-SNAPSHOT.jar" + os.pathsep + d + "javax.json-1.0.4.jar"


def class_path_chocosolver():
    d = directory_of_solver("chocosolver")
    return d + "choco-parsers-4.10.3-SNAPSHOT-jar-with-dependencies.jar"


class Instantiation:
    def __init__(self, pretty_solution, variables, values):
        self.pretty_solution = pretty_solution
        self.variables = variables
        self.values = values

    def __repr__(self):
        return self.variables, self.values

    def __str__(self):
        return str(self.pretty_solution)


class SolverPy4J:
    gateways = []
    processes = []

    def __init__(self, *, name, command):
        self.gateway, self.process = SolverPy4J.connexion(command)
        SolverPy4J.gateways.append(self.gateway)
        SolverPy4J.processes.append(self.process)
        self.solver = self.gateway.entry_point.getSolver()
        self.name = name

    @staticmethod
    def connexion(command):
        process = subprocess.Popen(command.split())
        cnt = 0
        while True:
            time.sleep(0.1)
            cnt += 1
            print("Py4J Connection " + str(cnt) + " ...")
            try:
                gateway = JavaGateway(eager_load=True)
            except Py4JNetworkError:
                print("Py4J Connection failed: No JVM listening ...")
            else:
                print("Py4J Successfully connected to the JVM")
                return gateway, process
        return gateway, process

    @staticmethod
    def close():
        for element in SolverPy4J.gateways:
            element.close()

    def loadXCSP3(self, arg):
        if isinstance(arg, str):
            self.solver.loadXCSP3(arg)
        elif isinstance(arg, IOBase):
            self.solver.loadXCSP3(arg.name)


class SolverProcess:
    def __init__(self, *, name, command):
        self.name = name
        self.command = command
        self.stdout = None
        self.stderr = None

    def solve(self, model, time_limit=0, n_restarts=0):
        stopwatch = Stopwatch()
        print("\nSolving by " + self.name + " in progress ... ")
        if time_limit != 0:
            model += " -t=" + str(time_limit) + "s"
        if n_restarts != 0:
            model += " -r_n=" + str(n_restarts)
        model += " -cm -valh=Last"
        print("  command: ", self.command + " " + model)
        result = self.execute(self.command + " " + model)
        print("Solved by " + self.name + " in " + stopwatch.elapsed_time() + " seconds (add the option -ev to see totally the output of the solver).")
        return self.solution() if result else None

    def execute(self, command):
        try:
            p = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, error = p.communicate()
            self.stdout, self.stderr = out.decode('utf-8'), error.decode('utf-8')
            return True
        except KeyboardInterrupt:
            out, error = p.communicate()
            self.stdout, self.stderr = out.decode('utf-8'), error.decode('utf-8')
            print(self.stdout)
            return False

    def solution(self):
        if self.stdout.find("<unsatisfiable") != -1 or self.stdout.find("s UNSATISFIABLE") != -1:
            return Instantiation("unsatisfiable", None, None)
        if self.stdout.find("<instantiation") == -1 or self.stdout.find("</instantiation>") == -1:
            print("  actually, the instance was not solved (add the option -ev to have more details")
            if options.ev:
                print("\n", self.stdout)
            return None
        left, right = self.stdout.find("<instantiation"), self.stdout.find("</instantiation>")
        root = etree.fromstring(self.stdout[left:right + len("</instantiation>")], etree.XMLParser(remove_blank_text=True))
        variables = []
        for token in root[0].text.split():
            r = VarEntities.get_item_with_name(token)
            if isinstance(r, (EVar, Variable)):  # TODO why do we need these two cases?
                variables.append(r)
            else:
                for x in flatten(r.variables, keep_none=True):
                    variables.append(x)
        values = root[1].text.split()  # a list with all values given as strings (possibly '*')
        assert len(variables) == len(values)
        for i, v in enumerate(values):
            if variables[i]:
                variables[i].value = v  # we add new field (may be useful)

        pretty_solution = etree.tostring(root, pretty_print=True, xml_declaration=False).decode("UTF-8").strip()
        return Instantiation(pretty_solution, variables, values)



from pycsp3 import *

x, y, z = VarArray(size=3, dom={0, 1})

try:
    satisfy(Sum([1, 2, 3]) < 5)
except TypeError as e:
    print("Test assert terms sum parameter 1:", e)

try:
    satisfy(Sum(None) < 5)
except TypeError as e:
    print("Test assert terms sum parameter 2:", e)

satisfy(Sum(x) < 5)

try:
    satisfy(Sum(x, [1]) < 10)
except TypeError as e:
    print("Test assert terms sum parameter 3:", e)

satisfy(Sum(x, y) < 10)

try:
    satisfy(Sum(x))
except AssertionError as e:
    print("Test assert partial constraint:", e)

try:
    satisfy(Sum(x) > None)
except TypeError as e:
    print("Test assert type partial constraint 1:", e)

try:
    satisfy(Sum(x) > {1, 2, 3})
except TypeError as e:
    print("Test assert type partial constraint 2:", e)

satisfy(Sum(x) in {1, 2, 3})

try:
    satisfy(Sum(x) in [1, 2, 3])
except AssertionError as e:
    print("Test assert bad partial constraint:", e)

try:
    satisfy(Sum(x, condition=[]))
    satisfy(Sum(x, condition=(0)))
except AssertionError as e:
    print("Test assert condition in keyword argument:", e)

satisfy(Sum(x, condition=("EQ", 0)))
satisfy(Sum(x, condition=["GE", 0]))
